#!/usr/bin/env perl
# 
# Copyright (c) 2009-2023 The TEBA Project. All rights reserved.
# 
# Redistribution and use in source, with or without modification, are
# permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# 
# Author: Atsuhi Yoshida


use warnings;
use strict;

use FindBin qw($Bin);
use lib "$Bin/../TEBA";

use JSON;

use Data::Dumper;
use Carp qw(confess);
$SIG{__DIE__} = \&confess;

use CFG;

# JSON 形式の AST を読み込む
my $root = JSON->new->decode(join('', <>));

# 読み込んだ AST から expression level の CFG オブジェクトを生成する
my $cfg = CFG->new_expr($root);

# CFG 上を動く visitor を作成。解釈実行するので、順方向に移動する。
my $visitor = $cfg->visitor->set_direction_next;

# visitor に計算のための stack と、変数の表を作成
# visitor 間で共有する。
$visitor->{stack} = CFG::STACK->new;
$visitor->{var} = my $var = { 'printf:#GLOBAL' => '#printf' };

# 各ノードでの処理を行う手続きを登録
$visitor->build(&analyzer_tbl);

#$visitor->{env_sig} = undef;
#$visitor->{clone_copy} = undef;
# visitor は、繰返しでは同じノードを何度も通過するので、訪問記録は不要。
# よって、env_sig は undef のままにする。
# stack と var は、visitor 間で共有するので、shallow copy されればよい。
# 複製のための特別な処理は不要なのえ、clone_copy も undef のままにする。

# 開始ノードに移動し、実行
foreach ($cfg->node_begin) {
    $visitor->move_to($_);
    $visitor->run;
    print Dumper($var); # 変数の表を出力
    print Dumper($visitor->{stack}); # スタックを出力
}

# 解釈処理の表
sub analyzer_tbl {
    return [
	{
	    p => "literal", # 定数値
	    a => sub {
		my $v = shift;
		my $value = $v->ast_node->attr('value');
		$value =~ s/\\\\/\\/g if $value =~/^"/; # 文字列の場合のエスケープ処理
		$v->{stack}->push($value);
	    }
	},
	{
	    p => "rvalue", # 変数の値の参照(右辺値)
	    a => sub {
		my $v = shift;
		$v->{stack}->push($v->{var}->{$v->ast_node->uniq_varname} || 0);
		# 初期化されていないときは 0 にする
	    }
	},
	{
	    p => "lvalue", # 左辺値 (変数名を使用)
	    a => sub {
		my $v = shift;
		$v->{stack}->push($v->ast_node->uniq_varname);
	    }
	},
	{
	    p => "lrvalue", # increment など、左辺値と右辺値の両方がある場合
	    a => sub {
		my $v = shift;
		$v->{stack}->push($v->ast_node->uniq_varname); # 右辺値を push
		$v->{stack}->push($v->{var}->{$v->ast_node->uniq_varname} || 0); # 左辺値を push
		# 右辺値だけを push する方法もあるが、左辺値も push すると、
		# 代入の処理のときに、必ず2つの値があることになり、処理を統一的に書ける(はず)。
	    }
	},
	{
	    p => "assign", # 代入の場合
	    a => sub {
		my $v = shift;
		my $sym = $v->ast_node->op_symbol;
		my $rvalue = $v->{stack}->pop;  # 右辺の値
		my $var = $v->{stack}->pop;     # 左辺の変数
		if ($sym eq "_=_") {  # 普通の代入
		    $v->{var}->{$var} = $rvalue;
		    $v->{stack}->push($v->{var}->{$var}); # 左辺の変数の値が全体の評価値
		} elsif ($sym eq "_++") { # 後置 increment の場合
		    $v->{stack}->push($rvalue);  # increment 前の値が全体の評価値
		    $v->{var}->{$var} = $rvalue + 1; # 変数の値を1増やす
		} else {
		    die "Not supported yet: ".$v->ast_node->op_symbol.".";
		}
	    }
	},
	{
	    p => "op", # 代入以外の演算子
	    a => sub {
		my $v = shift;
		my $sym = $v->ast_node->op_symbol;
		my @val = ($v->{stack}->pop, $v->{stack}->pop);
		if ($sym eq "_+_") {
		    $v->{stack}->push($val[1] + $val[0]);
		} elsif ($sym eq "_-_") {
		    $v->{stack}->push($val[1] - $val[0]);
		} elsif ($sym eq "_>_") {
		    $v->{stack}->push($val[1] > $val[0] ? 1 : 0);
		} elsif ($sym eq "_<_") {
		    $v->{stack}->push($val[1] < $val[0] ? 1 : 0);
		} else {
		    die "Not supported yet: $sym.";
		}
		# ここでは真面目に各演算子の処理を書いているが、perl の eval 関数を使えば
		# まとめて一気に処理できる。
		#   $sym =~ s/_/$val[1]/;
		#   $sym =~ s/_/$val[2]/;
		#   $v->{stack}->push(eval($sym);
	    }
	},
	{
	    p => "call", # 関数呼出し
	    a => sub {
		my $v = shift;
		my @parts = split('_', $v->label);
		my @args = ();
		while (@parts > 2) {
		    pop(@parts);
		    unshift(@args, $v->{stack}->pop);
		}
		my $func = $v->{stack}->pop;
		if ($func =~ '^#') { # Perl に直接対応する関数あり
		    $func =~ s/^#//;
		    my $call = sprintf("(%s(%s))", $func, join(',', @args));;
		    eval($call);
		    warn ("Error func call \"$call\" ", $@) if $@;
		}
	    }
	},
	{
	    p => qr/^#(expr_st|for_(init|succ))-end$/,
	    # 式文、for 文の初期化式、前進式の計算結果がスタックに残っているので、削除
	    a => sub {
		my $v = shift;
		$v->{stack}->pop;
	    }
	},
	{
	    p => "#true", # 分岐で真の場合に進んだら
	    a => sub {
		my $v = shift;
		my $val = $v->{stack}->pop;
		unless ($val) {  # 偽ならこの visitor は破棄
		    $v->move_to_final_state;
		    $v->{stack}->push($val); # stack を戻す
		} else {          # 真ならこの visitor は継続するので、他の visitor を破棄
		    $v->kill_queue;
		}
	    }
	},
	{
	    p => "#false", # 分岐で偽の場合に進んだら
	    a => sub {
		my $v = shift;
		my $val = $v->{stack}->pop;
		if ($val) {  # 真ならこの visitor は破棄
		    $v->move_to_final_state;
		    $v->{stack}->push($val); # stack を戻す
		} else {          # 偽ならこの visitor は継続するので、他の visitor を破棄
		    $v->kill_queue;
		}
	    }
	}
	];
}
